!
! Version 30.04.2002 - H. Forbert   harald.forbert@theochem.ruhr-uni-bochum.de
! $Id: fourier_sums.f90,v 1.1 2008/06/30 20:48:37 akohlmey Exp akohlmey $
!
MODULE fourier_sums
  
  USE kinds, ONLY : dbl

  IMPLICIT NONE

  PRIVATE
  PUBLIC :: fourier_sum
  PUBLIC :: direct_fourier_sum

CONTAINS

!! calculate cos_sum =  sum_j cos(j*w) input_j
!!       and sin_sum =  sum_j sin(j*w) input_j
!!
!! sums start at 0, but indices of input start at 1, e.g.
!! cos_sum = sum_{j=1}^n cos((j-1)*w) input_j

SUBROUTINE fourier_sum ( n, input, omega, f_sum )

  INTEGER, INTENT(IN) :: n
  REAL (dbl), DIMENSION(:), INTENT(IN) :: input
  REAL (dbl), INTENT(IN) :: omega
  REAL (dbl), INTENT(OUT) :: f_sum

  ! using the algorithm by Reinsch, as described in
  ! "Numerische Mathematik 1" by J. Stoer, 5th edition, Chapter 2.3.2
  ! page 69f.

  REAL (dbl) :: lambda, duk, uk, cs, ss, cf, sf
  INTEGER    :: k

  ! in order to be able to sum up the input_j in ascending order
  ! use above algorithm with inverse data ordering and correct
  ! omega -> - omega and the new origin at the end

  uk = 0.0_dbl
  duk = 0.0_dbl
  IF (COS(omega) > 0.0_dbl) THEN
     lambda = -4.0_dbl*SIN(0.5_dbl*omega)**2
     DO k = 1, n
        uk  = uk + duk
        duk = lambda*uk + duk + input(k)
     END DO
  ELSE                                 ! cos(omega) <= 0.0_dbl
     lambda = 4.0_dbl*cos(0.5_dbl*omega)**2
     DO k = 1, n
        uk  = duk - uk
        duk = lambda*uk - duk + input(k)
     END DO
  END IF
  cs = duk - 0.5_dbl * lambda * uk
  ss = uk * SIN(omega)

  ! now correct for ordering:
  cf = COS(omega*(n-1))
  sf = SIN(omega*(n-1))

  ! final result
  f_sum = cf*cs*cf*cs+sf*ss*sf*ss+sf*cs*sf*cs+cf*ss*cf*ss

  RETURN

END SUBROUTINE fourier_sum

!! calculate cos_sum =  sum_j cos(j*w) input_j
!!       and sin_sum =  sum_j sin(j*w) input_j
!!
!! sums start at 0, but indices of input start at 1, e.g.
!! cos_sum = sum_{j=1}^n cos((j-1)*w) input_j

SUBROUTINE direct_fourier_sum ( n, input, omega, cos_sum, sin_sum )

  INTEGER, INTENT(IN) :: n
  REAL (dbl), DIMENSION(:), INTENT(IN) :: input
  REAL (dbl), INTENT(IN) :: omega
  REAL (dbl), INTENT(OUT) :: cos_sum
  REAL (dbl), INTENT(OUT) :: sin_sum

  INTEGER :: i

  cos_sum = 0.0_dbl
  sin_sum = 0.0_dbl
  DO i = 1, n
    cos_sum = cos_sum + COS((i-1)*omega) * input(i)
    sin_sum = sin_sum + SIN((i-1)*omega) * input(i)
  END DO

  RETURN
   
END SUBROUTINE direct_fourier_sum

END MODULE fourier_sums
