
PROGRAM fourier

!
! This program takes an input file with time-data and calculates the
! powerspectrum/fourier transform of the autocorrelation function.
! The input file must only have lines of the form
!
! <integer> <dummy 1> <dummy 2> <dummy 3> <data 1> <data 2> <data 3>
!
! where the first four entries are ignored and the autocorrelation/
! fourier-transform is performed on the last 3 entries of a line.
! The name of the input file as well as other needed quantities
! are queried by the program interactively.
!
! The output file will contain the power-spectrum up to a maximum frequency.
! The format of this file is:
!
! <frequency> <value1> <value2> <value3> <value4> <value5>
! 
! <frequency> is the frequency in wave numbers.
! <value1>    is the plain power spectrum of the input data (normalized to
!             unity in the output frequency range.
! <value2>    is the power spectrum with a prefactor of 
!             \omega ( 1 - \exp(-\beta \hbar \omega) )
!             corresponding to the classical/Gordon limit.
! <value3>    is the power spectrum with a prefactor of
!             \omega \tanh(\beta \hbar \omega/2)
!             corresponding to the Kubo correction
! <value4>    is the power spectrum with a prefactor of
!             \omega \beta \hbar \omega
!             corresponding to the high temperature / harmonic limit
!             Note: this is the recommended correction factor.
! <value5>    is the power spectrum with a prefactor of
!             \omega ( 1 - \exp(-\beta \hbar \omega) ) *
!                                              \exp(\beta \hbar \omega /2)
!             corresponding to Schofield's correction
! All spectra with their corresponding prefactor are separately normalized
! in the output range to sum up to unity.
!
! Note: the index of refraction of the medium is set to unity.
!
! Version 30.4.2002 - H. Forbert   harald.forbert@theochem.ruhr-uni-bochum.de
! $Id: main.f90,v 1.1 2008/06/30 20:48:44 akohlmey Exp akohlmey $
!
  USE read_data, ONLY  : check_file, read_file
  USE write_data, ONLY : write_file
  USE kinds, ONLY : dbl
  USE fourier_sums, ONLY : fourier_sum

  IMPLICIT NONE

  INTEGER :: n,nn,i,k,ios, nstep, nover, nthreads
  INTEGER, EXTERNAL :: OMP_GET_NUM_THREADS
  CHARACTER (len=255) :: in_filename, out_filename
  REAL (dbl), ALLOCATABLE, DIMENSION ( : , : ) :: input_data
  REAL (dbl), ALLOCATABLE, DIMENSION ( : ) :: ftrans
  REAL (dbl), ALLOCATABLE, DIMENSION ( : ) :: wtrans

  REAL (dbl) :: t, f, c, e, s, pi, dt, f1, f2, f3
  REAL (dbl) :: data_t,data_temp,max_wave
  REAL (dbl) :: wave_fac, bh, avg(3)

  pi = ACOS(-1.0_dbl)

  print*,"Fourier version 06/2008"
!
! OpenMP info
!$OMP parallel
!$ nthreads = OMP_GET_NUM_THREADS()
!$OMP master
!$ print*,'OpenMP version using ', nthreads, ' thread(s).'
!$OMP end master
!$OMP end parallel
!
  PRINT *,"give time-data input filename (CPMD DIPOLE format file):"
  READ (*,'(A255)') in_filename

  PRINT *,"give output filename:"
  READ (*,'(A255)') out_filename

  PRINT *,"give time difference in input file (in atomic units):"
  PRINT *,"[CAUTION: this is the MD time step times the sampling frequency.]"
  READ (*,*) data_t
  wave_fac = 219474.0_dbl/data_t

  PRINT *,"give temperature (in Kelvin):"
  READ (*,*) data_temp
  bh = 1.05459e-34_dbl/1.38066e-23_dbl/2.41889e-17_dbl/data_t/data_temp

  PRINT *,"Output Max Frequency (Wavenumbers):"
  READ (*,*) max_wave

  PRINT *,"give integer frequency increment for output"
  PRINT *,"1 = finest grid/direct fourier transform - standard deviation 100%"
  PRINT *,"n = every n-th frequency - standard deviation about 1/sqrt(n)"
  PRINT *,"Larger numbers result in a coarse but less noisy spectrum"
  READ (*,*) nstep

  IF (nstep < 1) THEN
     PRINT *,"NSTEP must be bigger or equal 1."
     STOP
  END IF

  nover = nstep

  CALL check_file(in_filename,n,avg)

  ALLOCATE( input_data ( n+1 , 3 ), STAT = ios)
  IF (ios /= 0) THEN
     PRINT *,"Out of memory, while trying to allocate array 'input_data'."
     STOP
  END IF

  CALL read_file(input_data,in_filename,n,avg)

  nn = INT(n*max_wave/wave_fac/(2.0_dbl*pi))
  IF (nn+1 > n) THEN
     PRINT *,"Maximum wave number too big"
     STOP
  END IF
  nn = nn/nstep
  PRINT *,"Number of frequencies: ",nn

  ALLOCATE( ftrans ( nn+1 ), STAT = ios)
  IF (ios /= 0) THEN
     PRINT *,"Out of memory, trying to allocate 'ftrans'."
     STOP
  END IF

  ALLOCATE( wtrans ( nn + 1 ), STAT = ios)
  IF (ios /= 0) THEN
     PRINT *,"Out of memory, trying to allocate 'wtrans'."
     STOP
  END IF

! Welch windowing function:
! Welch function zero at indices 0 and n+1
  input_data ( n+1, : ) = 0.0_dbl

!$OMP parallel do private(i) schedule(static)
  DO i = 1, n
    input_data(i,:) = input_data(i,:) * (1.0_dbl - ((2*i-n-1.0_dbl)/(n+1))**2)
  END DO

  dt = 2.0_dbl*nstep*pi/((n+1)*nover)

!$OMP parallel do private(i,k,t,c,f,f1,f2,f3,s,e) schedule(static)
  DO i = 0, nn

!    Base frequency
     t= 2.0_dbl*i*nstep*pi/(n+1)
     c = 0.0_dbl

     DO k = 0, nover-1
        f = 0.0_dbl
        CALL fourier_sum(n,input_data(:,1),t+k*dt,f1)
        CALL fourier_sum(n,input_data(:,2),t+k*dt,f2)
        CALL fourier_sum(n,input_data(:,3),t+k*dt,f3)
        f = f1+f2+f3

!       input data should have zero mean...
        IF (i+k == 0) f = 0.0_dbl

!       apply cubic spline correction for input data
!       see for example
!       "Numerische Mathematik 1" by J. Stoer, 5th edition, Chapter 2.3.4
!       page 81.

        s=0.5_dbl*(t+k*dt)
        IF (s > 0.1) THEN
          e=(SIN(s)/s)**4
        ELSE
          e=(1.0_dbl-s**2/6.0_dbl+s**4/120.0_dbl)**4
        END IF

        e = e*3.0_dbl/(1.0_dbl+2.0_dbl*COS(s)**2) 
        c = c + e**2*f

     END DO

!    store result

     wtrans(1+i) = t+0.5_dbl*dt*(nover-1)
     ftrans(1+i) = c

  END DO

! write file with different normalizations.
  PRINT *,"writing output to: ", out_filename
  CALL write_file(ftrans,wtrans,out_filename,nn,bh,wave_fac)
  PRINT *,"done."

  STOP

END PROGRAM fourier
