      PROGRAM trajec2atoms
      IMPLICIT NONE
c     Written by Robert Williams, Uniformed Services University,
c     4301 Jones Bridge Road, Bethesda, MD 20814
c     rwilliams@usuhs.edu or bob@bob.usuhs.mil
c
c     $Id: trajec2atoms.f,v 1.1 2008/06/30 20:48:26 akohlmey Exp akohlmey $
c     
c     This program converts the CPMD TRAJECTORY output, which contains
c     coordinates and velocities for all N atoms at each step of a 
c     molecular dynamics trajectory, into N separate TRAJEC.00n files, 
c     each containing the coordinates and velocities for a particular
c     atom at each step of the MD trajectory.
c     This code also attempts to correct for the missing step error
c     seen in some BOMD restarts, where the new TRAJECTORY data is not
c     appended continuously to the old TRAJECTORY file, by performing 
c     a simple linear interpolation between the steps bracketing the 
c     missing step.   
c     
c     The code is written is a conservative Fortran 77 style
c     and should compile using f2c or any subsequent open compiler.c 

c     maxatoms = dimension of cart(maxatoms,3), atomnumber(maxatoms),
c     atoms(maxatoms)
c     maxbdim = 3*maxatoms = row and column dimensions of b.
c     b(maxbdim,maxbdim) is the cartesian coordinate force matrix.
c     cart(maxatoms,3) = cartesian coordinates

      INTEGER maxatoms
      PARAMETER (maxatoms=1024)  

      DOUBLE PRECISION cart(maxatoms,3), velocity(maxatoms,3)
      DOUBLE PRECISION cart_prev(maxatoms,3), velocity_prev(maxatoms,3)
      DOUBLE PRECISION cart_interp(3), velocity_interp(3)

      INTEGER nstep, istep, iatom, iiatom, icount, iunit_base, lun, ilun
      INTEGER natoms, ncount, ncount_previous, nprevious_step, modulo
      INTEGER k, kk, kkk, mm, ijstep, nback, iback, jatom, nthisstep

      CHARACTER inpfile*256,outfile*256,card*150, ext*3

c     1         2         3         4         5         6         7         8
c     2345678901234567890123456789012345678901234567890123456789012345678901234567890
c     ----------------------------------------------------------------------

      WRITE(*,*)'trajec2atoms converts a CPMD TRAJECTORY file for N'
      WRITE(*,*)'atoms into N TRAJEC.00n files, each containing the'
      WRITE(*,*)'coordinates and velocities for one specific atom.'
      WRITE(*,*)'These files are appropriate input for the fourier'
      WRITE(*,*)'program written by H. Forbert and modified by '
      WRITE(*,*)'Axel Kohlmeyer for calculating the fourier transform'
      WRITE(*,*)'of the velocity autocorrelation for a CPMD DIPOLE'
      WRITE(*,*)'file.  The DIPOLE file contains the velocities of'
      WRITE(*,*)'the dipole moment trajectory.'
c     72 
c     This block provides optional argument line input of the filename.
      CALL GETARG(1,inpfile)
      IF ((inpfile(1:1).EQ.' ').OR.(inpfile(1:1).EQ.'-')) THEN
C     No argument to command line. Prompt for the data filename.
        WRITE(*,"('Input filename [TRAJECTORY]: ',$)")
        READ (*,'(A255)') inpfile
        IF ((inpfile(1:1).EQ.' ').OR.(inpfile(1:1).EQ.'-')) THEN
          WRITE(inpfile,"('TRAJECTORY')")
        ENDIF
      ENDIF

c     The number of atoms could be obtained by reading the first
c     natoms+1 lines of the trajectory and noting where the step
c     number changes.
      ncount_previous=10000000
      OPEN(9,FILE=inpfile,STATUS='OLD',FORM='FORMATTED',ERR=999)
      DO icount = 1,10000000
c     Read a card from the TRAJECTORY file until the step
c     number changes.
        READ(9,'(i8)',end=666,err=666) ncount
        IF ( (icount .GT. 1) .AND. ( ncount .GT. ncount_previous))
     &       THEN
          natoms = icount - 1
          PRINT *, 'Found ',natoms,' atoms.'
          IF (natoms.gt.999) STOP 'can handle only up to 999 atoms'
c     should change this to rewind...
          CLOSE (9)
          GO TO 889
        ENDIF
        ncount_previous = ncount
      END DO
 889  CONTINUE

      print *,'Re-Opening input file:  ',inpfile(:index(inpfile,' '))
      OPEN(9,file=inpfile,status='old',form='formatted')
      PRINT *,'Writing output for ',natoms,' atoms'
      PRINT *,'in the form: TRAJEC:001, TRAJEC.002, ...',natoms
      PRINT *,'Working....This may take a while.'

ccccccccccccccccccccccccccccccccccccccccccccccccc
cccccccccccccccopen output files ccccccccccccccc
************open file for each atom ************
      
      iunit_base = 9
      DO iatom = 1, natoms

c     Make the output filename extension.
        IF(iatom.LT.10) THEN
          WRITE(ext,'(''00'',i1)') iatom
        ELSEIF(iatom.gt.99) THEN
          WRITE(ext,'(i3)') iatom
        ELSE
          WRITE(ext,'(''0'',i2)') iatom
        ENDIF
        outfile = 'TRAJEC.'//ext  
        lun = iunit_base + iatom
        OPEN(lun,file=outfile,status='unknown',form='formatted')
      END DO
c     All output files are now open.
ccccccccccccccccccccccccccccccccccccccccccccccccc
cccccccccdone opening output files ccccccccccccc


      modulo = 5000
      nprevious_step = 0
c     Assume ten million steps and bail out after last read card.       
      DO istep=1,10000000
        DO iatom=1,natoms
c     Read a card from the TRAJECTORY file and bail out
c     if the end is reached.
          READ(9,FMT='(a)',END=666,ERR=668) card
          READ(card,FMT=444,ERR=74)  nstep,       
     &         (cart(iatom,k),k=1,3),(velocity(iatom,k),k=1,3)
 444      FORMAT(i7,6D24.14)

          GO TO 75
 74       CONTINUE
          PRINT *,'Error reading card at nstep = ',nstep
          PRINT *,'and atom number ',iatom,'.  Card = ' 
          PRINT *,card

c     "   <<<<<<  NEW DATA  >>>>>>" lines mark the beginning of
c     restarted trajectories.   If a NEW DATA line is found,
c     read lines until nstep is one greater than nprevious_step. 
          IF ( INDEX(card, 'NEW DATA').GT.0 ) then
            PRINT *,'Found a NEW DATA line.  Compensating.'
c     We don't expect more than 100000 overlap lines, do we?
            DO kkk = 1,100000
              READ(9,FMT='(a)',END=666,ERR=668) card
              READ(card,444,ERR=76)  nstep,       
     &             (cart(iatom,k),k=1,3),(velocity(iatom,k),k=1,3)
              IF ( nstep .EQ. (nprevious_step + 1) ) THEN
c     Back up one line, jump out of this loop, and
c     continue reading.
                PRINT *,'Found next logical step at ',NSTEP,'.'
                PRINT *,'Continuing....'
                BACKSPACE (9)
                GO TO 9992
              ELSE IF ( nstep .LT. nprevious_step ) THEN
                PRINT *,'Found an overlap.  nstep = ',nstep
                PRINT *,'nprevious_step = ',nprevious_step
                PRINT *,'Attempting to join the restart files'
                PRINT *,'at step number ',nstep
                nback = nprevious_step - nstep + 1

c     print *,'Backspacing atom files ',nback,' lines.'
                DO iback = 1, nback
                  DO jatom = 1, natoms
                    ilun = iunit_base + jatom
                    backspace (ilun)
                  END DO
                END DO
                backspace (9)
                nprevious_step = nstep - 1
                GO TO 9992

              ELSE 
c     if ( nstep .gt. (nprevious_step +1) ) then
c     Hmmmm dont know what to do about this yet.
                PRINT *,'Skip found after NEW DATA LINE.'
                PRINT *,'Check your TRAJECTORY file.  Stopping.'
                STOP
              ENDIF
            END DO

            GO TO 77
 76         CONTINUE
            PRINT *,'Error reading card at nstep = ',nstep
            PRINT *,'after a NEW DATA line.   Card = ' 
            PRINT *,card
            PRINT *,'You will have to split and edit your'
            PRINT *,'TRAJECTORY file to see what is wrong.'
            PRINT *,'Stopping.'
            STOP
 77         CONTINUE

          ENDIF
          STOP
 75       CONTINUE


c     Keep track of the step number.  There should be
c     natoms of the same step number.  Step numbers in
c     the trajectory file may not start at 1. 
          IF (iatom .EQ. 1) nthisstep = nstep
c     Now check every step in this iatom loop.
          IF (nthisstep .NE. nstep) THEN
            PRINT *,'Sequence error!!!  '
            PRINT *,'This step should be ', nthisstep, ','
            PRINT *,'but I read step ', nstep
            PRINT *,'at the following line:'
            PRINT *,card
            PRINT *,'Stopping.  Check your TRAJECTORY file.'
            STOP
          ENDIF

c     Also keep track of the previous step number so that
c     we can fix the skipped step number problem in bomd
c     trajectory files in cpmd 3.9.2 and early devel versions
c     of 4.0.  


          IF ( (iatom .EQ. 1) .AND. (istep .GT. 1) .AND.
     &         (nstep .NE. (nprevious_step + 1) ) )  THEN
c     A bomd restart has skipped a step.  
c     Use the saved previous step and this step to interpolate.                  
            PRINT *,'Found a skipped step for atom ',iatom,
     &           'between steps ',nprevious_step,' and ',nthisstep
            PRINT *,'Fixing it with a linear interpolation.'
            PRINT *,'Check output for the word "interpolated"'
            PRINT *,'at the end of the interpolated line.'
            DO iiatom = 1, natoms
              DO kk = 1,3
                cart_interp(kk) = ( cart_prev(iiatom,kk) 
     &               + cart(iiatom,kk) ) / 2.0d0
                velocity_interp(kk) = ( velocity_prev(iiatom,kk)
     &               + velocity(iiatom,kk) ) / 2.0d0
c     72 
              END DO
              lun = iunit_base + iiatom
c     Write the interpolated step for each atom to the appropriate file.
              WRITE(lun,445)  (nprevious_step + 1),       
     &             (cart_interp(mm), mm=1,3),
     &             (velocity_interp(mm), mm=1,3)
 445          FORMAT(i6,6F24.14,'          interpolated')
            END DO

c     At the end of the skipped set, reset nstep
c     so that nprevious step is correctly set a few
c     lines below.
            IF (iatom .EQ. natoms) THEN
              nstep = nprevious_step + 1
            ENDIF
          ENDIF


c     nprevious_step will equal nstep while reading lines for
c     atoms in the same step.

          nprevious_step = nstep

c     Now write the step for each atom to the appropriate output file.
          lun = iunit_base + iatom
          WRITE(lun,'(a)') card
        END DO

c     Save the previous step data. 
        DO iatom = 1,natoms
          DO k = 1,3
            cart_prev(iatom,k) = cart(iatom,k)
            velocity_prev(iatom,k) = velocity(iatom,k)
          END DO
        END DO


c     Here is where we jump to after finding a "NEW DATA" line
c     and then reading the overlap steps until we find the
c     next step in the sequence.   We don't save the previous
c     data (above) because that was already saved.
 9992   CONTINUE

c     Print a dot every 5000 steps.
        ijstep = istep - modulo
        IF (ijstep .EQ. 0) THEN
          modulo = modulo + 5000
          WRITE (*,'(".",$)')
        ENDIF

      END DO

c     This is the bailout point for the end of trajectory file.
      GO TO 667
 666  CONTINUE
      PRINT *,'End of file found.  Successfully completed.' 
      STOP 'success'

 668  CONTINUE
      PRINT *,'Error reading card:' 
      PRINT *,card
 667  CONTINUE
      STOP 'ERROR'

 999  CONTINUE
      PRINT*,'error, exit'
      STOP
      END

c ************************************************************************
